#!/usr/bin/python

import json
import logging
import os

import requests
import urllib3
from ansible.module_utils.am2cm_commons.logging_utils import configure_logging
from ansible.module_utils.basic import AnsibleModule
from requests.auth import HTTPBasicAuth

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

configure_logging(os.path.basename(__file__).replace(".py", "-module.log"))
LOG = logging.getLogger(__name__)


def get_mappings_to_merge(params):
    """
    The goal of this function is to:
        - create a dict which looks like this:
          { <service_name>: { 'new_policy_id': <id>, 'legacy_policy_name': <name> }, ... }
        - filter output so that only those elements which contain both new and legacy keys (there may be services which
        only have one or the other, thus don't have to be merged)
    :param params: the module params coming from ansible
    :return: Mappings between new and legacy policies
    """
    legacy_to_new_policy_mapping_by_type = {}
    for service in params['all_ranger_repos']['services']:
        if service['type'] not in legacy_to_new_policy_mapping_by_type:
            legacy_to_new_policy_mapping_by_type[service['type']] = {}
        if "cm_" in service['name']:
            legacy_to_new_policy_mapping_by_type[service['type']]['new_policy_id'] = service['id']
        elif params['legacy_policy_prefix'] in service['name']:
            legacy_to_new_policy_mapping_by_type[service['type']]['legacy_policy_name'] = service['name']
    return dict((k, v) for (k, v) in legacy_to_new_policy_mapping_by_type.items() if len(v) > 1)


def run_module():
    module_args = dict(
        all_ranger_repos=dict(type='dict', required=True),
        legacy_policy_prefix=dict(type='str', required=True),
        ranger_load_balancer_url=dict(type='str', required=True),
        ranger_policy_url_path=dict(type='str', required=False, default='/service/plugins/policies/service/'),
        ranger_create_policy_url_path=dict(type='str', required=False, default='/service/public/v2/api/policy/apply'),
        ranger_admin_user=dict(type='str', required=False, default='admin'),
        ranger_admin_password=dict(type='str', required=True),
        verify_ssl=dict(type='bool', required=False, default=True)
    )

    result = dict(
        changed=False,
        original_message='',
        message=''
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )
    result['original_message'] = json.dumps(module.params)
    LOG.info("inputs: {}".format(result['original_message']))
    # ####################################################################
    # ####################################################################
    mappings_to_merge = get_mappings_to_merge(module.params)

    get_policy_url_prefix = module.params['ranger_load_balancer_url'] + module.params['ranger_policy_url_path']
    upsert_policy_url = module.params['ranger_load_balancer_url'] + module.params['ranger_create_policy_url_path']

    basic_auth = HTTPBasicAuth(module.params['ranger_admin_user'], module.params['ranger_admin_password'])
    verify = module.params['verify_ssl']
    params = {"mergeIfExists": "true"}

    for policy_type, policy_mapping in mappings_to_merge.items():
        new_policy_repo_url = get_policy_url_prefix + str(policy_mapping['new_policy_id'])
        new_policy_repo_response = requests.get(new_policy_repo_url, auth=basic_auth, verify=verify)
        new_policy_repo = json.loads(new_policy_repo_response.text)  # download new policy json by ID

        for policy in new_policy_repo['policies']:
            upsert_policy = dict(policy)  # copy downloaded new policy into tmp var
            upsert_policy['service'] = policy_mapping['legacy_policy_name']  # re-write field to legacy policy name
            upsert_response = requests.post(upsert_policy_url, json=upsert_policy,
                                            auth=basic_auth, verify=verify,
                                            params=params)  # POST it w/ mergeIfExists param
            if upsert_response.status_code == 200:
                result['message'] += f"Successfully added {policy['name']} policy " \
                                     f"from {policy['service']} to {upsert_policy['service']}\n"
            else:
                module.fail_json(msg=f"Couldn't add {policy['name']} policy "
                                     f"from {policy['service']} to {upsert_policy['service']}", **result)

    module.exit_json(**result)


def main():
    run_module()


if __name__ == '__main__':
    main()
