import json
import logging
import os
import time
from enum import Enum
from functools import partial

import cm_client
import requests
import urllib3
from ansible.module_utils.am2cm_commons.logging_utils import configure_logging
from cm_client import ApiService, ApiRoleConfigGroupList, ApiServiceConfig, \
    ApiCommand, ApiParcelList, ApiParcel, ApiConfigList, \
    ApiConfig, ApiCommandList
from cm_client.rest import ApiException

PARCEL_REPO_URLS = 'REMOTE_PARCEL_REPO_URLS'

urllib3.disable_warnings()

LOG = logging.getLogger(__name__)


class CDPCluster:
    """
    Class to represent a CDP Cluster.
    """

    DEFAULT_SERVICE_COMMAND_TIMEOUT_SECONDS = 900
    DEFAULT_ROLE_COMMAND_TIMEOUT_SECONDS = 900
    DEFAULT_RESTART_SERVICE_TIMEOUT_SECONDS = 900

    def __init__(self, hostname: str, port: int, is_https: bool, verify_ssl: bool,
                 username: str, password: str, cluster_name: str):
        self.hostname = hostname
        self.port = port
        self.protocol = 'https' if is_https else 'http'
        self.is_https = is_https
        self.verify_ssl = verify_ssl
        self.username = username
        self.password = password
        self.cluster_name = cluster_name

        self._set_configuration()

    def _set_configuration(self):
        LOG.debug("Setting Configuration object")
        configuration = cm_client.Configuration()
        api_version = self._determine_api_version()
        configuration.host = "{protocol}://{hostname}:{port}/api/{api_version}".format(protocol=self.protocol,
                                                                                       hostname=self.hostname,
                                                                                       port=self.port,
                                                                                       api_version=api_version)
        configuration.username = self.username
        configuration.password = self.password
        configuration.verify_ssl = self.verify_ssl

    def _determine_api_version(self):
        response = requests.get(url='{protocol}://{hostname}:{port}/api/version'.format(
            protocol=self.protocol, hostname=self.hostname, port=self.port),
            auth=(self.username, self.password),
            verify=self.verify_ssl)
        if not response.ok:
            LOG.error("Determining API version failed...")
            response.raise_for_status()

        LOG.info("API version: %s" % response.text.strip())
        return response.text.strip()

    def get_cm_version(self, fallback="0"):
        LOG.info("Trying to get the CM version, fallback value: {}".format(fallback))
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.get_version()
            LOG.debug("Response: {}".format(api_response))
            return api_response.version
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->get_version: %s\n" % e)
            return fallback

    def run_service_command(self, service_name: str, command_name: str,
                            timeout_sec: int = DEFAULT_SERVICE_COMMAND_TIMEOUT_SECONDS):
        LOG.info("Running service command '{command}' for service '{service}'...".format(command=command_name,
                                                                                         service=service_name))
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response: ApiCommand = api_instance.service_command_by_name(self.cluster_name, command_name,
                                                                            service_name)
            LOG.debug("Response: %s" % api_response)
            return self.wait_for_service_commands_to_run(service_name, command_name, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            return False, str(e)

    def get_parcel_info(self, parcel_product_name: str, parcel_version: str) -> ApiParcel:
        LOG.debug(f"Trying to get details of '{parcel_product_name}'")
        api_instance = cm_client.ParcelsResourceApi(cm_client.ApiClient())

        try:
            api_response: ApiParcelList = api_instance.read_parcels(self.cluster_name, view='full')
            for parcel in api_response.items:
                if parcel.product == parcel_product_name and parcel.version == parcel_version:
                    return parcel

            return None
        except ApiException as e:
            LOG.error("Exception when calling ParcelsResourceApi->read_parcels: %s\n" % e)
            raise e

    def set_parcel_repo(self, parcel_repo: str):
        LOG.debug(f"Trying to add '{parcel_repo}' parcel repo to CM")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            cm_configs = api_instance.get_config(view='full')
            old_parcel_repo_urls = None
            for cm_config in cm_configs.items:
                if cm_config.name == PARCEL_REPO_URLS:
                    old_parcel_repo_urls = cm_config.value
                    break

            # check if the repo has already been added
            for repo in old_parcel_repo_urls.split(','):
                if repo == parcel_repo:
                    LOG.info(f"'{parcel_repo}' has already been added to CM")
                    return

            # value is a comma-separated list
            new_parcel_repo_urls = old_parcel_repo_urls + ", " + parcel_repo
            new_cm_config = cm_client.ApiConfig(name=PARCEL_REPO_URLS, value=new_parcel_repo_urls)
            new_cm_configs = cm_client.ApiConfigList([new_cm_config])
            updated_cm_configs = api_instance.update_config(body=new_cm_configs)

            # wait to make sure parcels are refreshed
            time.sleep(10)

        except ApiException as e:
            LOG.error(f"Exception when adding {parcel_repo} parcel repo to CM: {e}\n")
            raise e

    def download_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.product, parcel_info.version)
        if current_info.stage != 'AVAILABLE_REMOTELY':
            LOG.info("Parcel is already downloaded, skipping...")
            return True, current_info

        try:
            api_instance.start_download_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting download of '{parcel_info.product}' polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.product, parcel_info.version).stage
                LOG.debug(f"Current stage of '{parcel_info.product}': '{current_stage}'")
                if current_stage == 'DOWNLOADED':
                    LOG.info("Download successful...")
                    success = True
                else:
                    time.sleep(30)

            return success, self.get_parcel_info(parcel_info.product, parcel_info.version)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_download_command: %s\n" % e)
            raise e

    def distribute_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.product, parcel_info.version)
        if current_info.stage != 'DOWNLOADED':
            LOG.info("Parcel is already distributed, skipping...")
            return True, current_info

        try:
            result: ApiCommand = api_instance.start_distribution_command(
                self.cluster_name, parcel_info.product, parcel_info.version)

            LOG.info(result)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting distribution of '{parcel_info.product}' polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.product, parcel_info.version).stage
                LOG.info(self.get_parcel_info(parcel_info.product, parcel_info.version))
                LOG.info(f"Current stage of '{parcel_info.product}': '{current_stage}'")
                if current_stage == 'DISTRIBUTED':
                    LOG.info("Distribution successful...")
                    success = True
                else:
                    time.sleep(30)

            return success, self.get_parcel_info(parcel_info.product, parcel_info.version)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_distribution_command: %s\n" % e)
            raise e

    def activate_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 1000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.product, parcel_info.version)
        if current_info.stage != 'DISTRIBUTED':
            LOG.info("Parcel is already activated, skipping...")
            return True, current_info

        try:
            api_instance.activate_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Activating parcel '{parcel_info.product}'.")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.product, parcel_info.version).stage
                LOG.info(f"Current stage of '{parcel_info.product}': '{current_stage}'")
                if current_stage == 'ACTIVATED':
                    LOG.info("Activation successful...")
                    success = True
                else:
                    time.sleep(10)

            return success, self.get_parcel_info(parcel_info.product, parcel_info.version)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->activate_parcel: %s\n" % e)
            raise e

    def undistribute_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.product, parcel_info.version)
        if current_info.stage != 'DISTRIBUTED':
            LOG.info("Parcel is not in distributed state, skipping...")
            return True, current_info

        try:
            api_instance.start_removal_of_distribution_command(
                self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Starting undistribution of '{parcel_info.product}'. Polling...")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.product, parcel_info.version).stage
                LOG.info(self.get_parcel_info(parcel_info.product, parcel_info.version))
                LOG.info(f"Current stage of '{parcel_info.product}': '{current_stage}'")
                if current_stage == 'DOWNLOADED':
                    LOG.info("Undistribution successful...")
                    success = True
                else:
                    time.sleep(15)

            return success, self.get_parcel_info(parcel_info.product, parcel_info.version)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->start_removal_of_distribution_command: %s\n" % e)
            raise e

    def deactivate_parcel(self, parcel_info: ApiParcel, timeout_sec: int = 5000):
        api_instance = cm_client.ParcelResourceApi(cm_client.ApiClient())

        current_info = self.get_parcel_info(parcel_info.product, parcel_info.version)
        if current_info.stage != 'ACTIVATED':
            LOG.info("Parcel is not in activated state, skipping...")
            return True, current_info

        try:
            api_instance.deactivate_command(self.cluster_name, parcel_info.product, parcel_info.version)

            success = False
            end_time = time.time() + timeout_sec
            LOG.info(f"Deactivating parcel '{parcel_info.product}'.")
            while time.time() < end_time and not success:
                current_stage = self.get_parcel_info(parcel_info.product, parcel_info.version).stage
                LOG.info(f"Current stage of '{parcel_info.product}': '{current_stage}'")
                if current_stage == 'DISTRIBUTED':
                    LOG.info("Deactivation successful...")
                    success = True
                else:
                    time.sleep(10)

            return success, self.get_parcel_info(parcel_info.product, parcel_info.version)
        except ApiException as e:
            LOG.error("Exception when calling ParcelResourceApi->deactivate_command: %s\n" % e)
            raise e

    def deploy_service_client_configuration(self, service_name: str):
        LOG.info("Deploying service client configuration for service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response: ApiCommand = api_instance.deploy_client_config_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)
            return True, api_response.result_message
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->deploy_client_config_command: %s\n" % e)
            return False, str(e)

    def start_service(self, service_name: str, timeout_sec: int = 360):
        LOG.info("Starting service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.start_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)

            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state(service_name, expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->start_command: %s\n" % e)
            raise e

    def stop_service(self, service_name: str, timeout_sec: int = 360):
        LOG.info("Stopping service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.stop_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)

            expected_service_state = {'service_state': 'STOPPED'}
            return self.wait_for_service_state(service_name, expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->stop_command: %s\n" % e)
            raise e

    def restart_service(self, service_name: str, timeout_sec: int = DEFAULT_RESTART_SERVICE_TIMEOUT_SECONDS,
                        healthchecks_to_ignore_if_their_state_is_concerning: list = []):
        LOG.info("Restarting service '%s'" % service_name)
        if service_name.lower() == 'mgmt':
            return self.restart_cm_management_service()

        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.restart_command(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)
            return self.wait_for_service_to_become_good_but_ignore_some_healthchecks(service_name, timeout_sec,
                                                                                     healthchecks_to_ignore_if_their_state_is_concerning)
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->restart_command: %s\n" % e)
            return False, str(e)

    def read_service(self, service_name: str):
        LOG.debug("Reading service '%s'" % service_name)
        try:
            if service_name == "mgmt":
                api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())
                api_response = api_instance.read_service()
            else:
                api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
                api_response = api_instance.read_service(self.cluster_name, service_name)
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->read_service: %s\n" % e)

    def is_service_exists(self, service_name: str):
        LOG.info(f"Checking whether {service_name} already exists on cluster {self.cluster_name}")
        try:
            return self.read_service(service_name=service_name) is not None
        except ApiException:
            return False

    def is_role_exists(self, service_name: str, role_name: str):
        LOG.info(f"Checking whether role '{role_name}' exists"
                 f" for service '{service_name}'"
                 f" on cluster {self.cluster_name}")
        try:
            roles: ApiRoleList = self.read_roles(service_name=service_name)
            next(role for role in roles.items if role.name == role_name)
            return True
        except StopIteration:
            return False

    def list_active_role_commands(self, role_name: str, service_name: str):
        LOG.debug("Listing active commands for role '%s' in service '%s'" % (role_name, service_name))
        api_instance = cm_client.RolesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.list_active_commands(self.cluster_name, role_name, service_name, view='summary')
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->list_active_commands: %s\n" % e)

    def list_active_service_commands(self, service_name: str):
        LOG.debug("Listing active commands for service '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.list_active_commands(self.cluster_name, service_name, view='summary')
            LOG.debug("Response: %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->list_active_commands: %s\n" % e)

    def list_active_cloudera_manager_resource_commands(self):
        """
        List active ClouderaManagerResource commands (e.g.: GenerateCredentials)

        :raises ApiException, If a connection issue occurred
        :return: ApiCommandList => A list of ApiCommands that are active currently.
        """
        LOG.debug("Listing active commands for cluster '%s'" % self.cluster_name)
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())
        try:
            return api_instance.list_active_commands(view='full')
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->list_active_commands: %s\n" % {e.reason})
            raise e

    def wait_for_service_state(self, service_name: str, expected_states: dict, timeout_sec: int = 360):
        """
        Waits until given service reaches the desired state defined by a dictionary or times out.
        See cm_client.models.api_service.ApiService implementation for attributes that can be checked.

        Example call - this will wait until the service_state is 'STARTED'
                        AND the health_summary is anything other than DISABLED:

            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            self.wait_for_service_state(service_name, expected_service_state, timeout_sec)

        :param service_name: The name of the service
        :param expected_states: Dictionary, where key is the attribute, value is the desired value.
            If value starts with '!', it means that we are expecting the given attribute to NOT equal to the given value
        :param timeout_sec: Maximum timeout in seconds
        :return: (True , message) if service reached the desired state before timeout (False, message) otherwise
        """
        success = False
        message = f"Service {service_name} did not reach desired state in {timeout_sec}s"
        begin_time = time.time()
        timeout = begin_time + timeout_sec
        while time.time() < timeout and not success:
            if service_name.lower() == 'mgmt':
                service_details: ApiService = self.read_cm_management_service()
            else:
                service_details: ApiService = self.read_service(service_name)
            all_succeeded = True
            for prop, value in expected_states.items():
                invert_logic = value.startswith('!')

                if not invert_logic:
                    if getattr(service_details, prop) != value:
                        all_succeeded = False
                        break
                else:
                    if getattr(service_details, prop) == value[1:]:
                        all_succeeded = False
                        break

            if all_succeeded:
                elapsed_seconds = time.time() - begin_time
                success = True
                message = f"Service {service_name} successfully reached desired state in {elapsed_seconds:.2f} seconds"
            else:
                time.sleep(10)

        return success, message

    def get_service_details(self, service_name: str) -> ApiService:
        return self.read_cm_management_service() if service_name.lower() == 'mgmt' else self.read_service(service_name)

    def service_can_be_considered_healthy(self, service_details, healthchecks_to_ignore_if_their_state_is_concerning):
        if service_details.health_summary == "GOOD":
            return True
        # if overall healthcheck is CONCERNING, we should see whether we can ignore
        # some of the CONCERNING healthchecks
        elif service_details.health_summary == "CONCERNING":
            # NOTE:
            # filtering logic should be read from bottom => top
            # i.e. the filter at the bottom gets executed first
            failed_healthchecks = list(
                # from the not GOOD healthchecks, the failed ones will be:
                #   - not CONCERNING (e.g. BAD)
                #   - are CONCERNING, but they are not on the ignore list either
                filter(lambda healtcheck: not (healtcheck.summary == "CONCERNING"
                                               and healtcheck.name
                                               in healthchecks_to_ignore_if_their_state_is_concerning),
                       # select anything that's not GOOD
                       filter(lambda healthcheck: healthcheck.summary != "GOOD",
                              # first select all enabled health checks
                              filter(lambda healthcheck: healthcheck.summary != "DISABLED",
                                     service_details.health_checks)
                              )
                       )
            )
            return not failed_healthchecks
        return False

    def wait_for_service_to_become_good_but_ignore_some_healthchecks(self, service_name: str, timeout_sec: int = 360,
                                                                     healthchecks_to_ignore_if_their_state_is_concerning: list = []):
        success = False
        message = f"Service {service_name} did not reach desired state in {timeout_sec}s. " + \
                  f"Ignored healthchecks: {healthchecks_to_ignore_if_their_state_is_concerning}"
        begin_time = time.time()
        timeout = begin_time + timeout_sec
        while time.time() < timeout and not success:
            service_details = self.get_service_details(service_name)

            if service_details.service_state == "STARTED" and self.service_can_be_considered_healthy(service_details,
                                                                                                     healthchecks_to_ignore_if_their_state_is_concerning):
                elapsed_seconds = time.time() - begin_time
                message = f"Service {service_name} successfully reached desired " \
                          f"state in {elapsed_seconds:.2f} seconds"
                success = True
                break

            time.sleep(10)

        return success, message

    def wait_for_commands_to_run(self, command_listing_function, command_name: str = None,
                                 timeout_sec: int = 360,
                                 success_message_function=lambda elapsed_seconds:
                                 f"Command has successfully finished in {elapsed_seconds:.2f} seconds",
                                 failure_message_function=lambda timeout:
                                 f"Command did not finish in {timeout} seconds"):
        """
        Waits until either all or at least the command specified by command_name has been executed.
        The way of getting a list of currently active commands depends on the command_listing_function argument

        :param command_listing_function: A function that returns with a list of active commands
        :param command_name: Command's name
        :param timeout_sec: Maximum timeout in seconds
        :param success_message_function: A function that return the message to return upon success
        :param failure_message_function: A function that return the message to return upon failure
        :raises ApiException, If command_listing_function has been unsuccessful contacting the server
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        success = False
        message = failure_message_function(timeout_sec)
        begin_time = time.time()
        timeout = begin_time + timeout_sec
        while time.time() < timeout and not success:
            active_commands: ApiCommandList = command_listing_function()
            # if the list of currently active commands is empty,
            # or it doesn't contain the command that we're waiting for, we're good to exit
            if not active_commands.items or (command_name and not list(command
                                                                       for command in active_commands.items
                                                                       if command.name == command_name)):
                elapsed_seconds = time.time() - begin_time
                success = True
                message = success_message_function(elapsed_seconds)
            else:
                time.sleep(10)
        return success, message

    def wait_for_role_commands_to_run(self, role_name: str, service_name: str, timeout_sec: int = 360):
        """
        Waits until all commands have been executed for a given role of a given service

        :param role_name: The exact name of the role (this should be queried beforehand from read_roles)
        :param service_name: The name of the service
        :param timeout_sec: Maximum timeout in seconds
        :raises ApiException, If couldn't connect to the server
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        return self.wait_for_commands_to_run(partial(self.list_active_role_commands, role_name, service_name),
                                             command_name=None,
                                             timeout_sec=timeout_sec,
                                             success_message_function=lambda elapsed_seconds:
                                             f"All commands for {role_name} in service {service_name} have successfully "
                                             f"finished in {elapsed_seconds:.2f} seconds",
                                             failure_message_function=lambda timeout:
                                             f"Commands for {role_name} in service {service_name} "
                                             f"did not finish in {timeout}s")

    def wait_for_service_commands_to_run(self, service_name: str, command_name: str, timeout_sec: int = 360):
        """
        Waits until all commands have been executed for a given service

        :param service_name: The name of the service
        :param command_name: The command to run on the service
        :param timeout_sec: Maximum timeout in seconds
        :raises ApiException, If couldn't connect to the server
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        return self.wait_for_commands_to_run(partial(self.list_active_service_commands, service_name),
                                             command_name,
                                             timeout_sec,
                                             success_message_function=lambda elapsed_seconds:
                                             f"All commands for service {service_name} have successfully "
                                             f"finished in {elapsed_seconds:.2f} seconds",
                                             failure_message_function=lambda timeout:
                                             f"Commands for service {service_name} did not finish in {timeout}s")

    def wait_for_cloudera_manager_resource_commands_to_run(self, command_name: str, timeout_sec: int = 360):
        """
        Waits until all commands have been executed for a given service

        :param service_name: The name of the service
        :param command_name: The command to run on the service
        :param timeout_sec: Maximum timeout in seconds
        :raises ApiException, If couldn't connect to the server
        :return: (True , message) if all commands have succeeded before timeout (False, message) otherwise
        """
        return self.wait_for_commands_to_run(partial(self.list_active_cloudera_manager_resource_commands),
                                             command_name,
                                             timeout_sec,
                                             success_message_function=lambda elapsed_seconds:
                                             f"All cluster commands have successfully "
                                             f"finished in {elapsed_seconds:.2f} seconds",
                                             failure_message_function=lambda timeout:
                                             f"Cluster Commands did not finish in {timeout}s")

    def wait_for_unnecessary_generate_credentials_commands(self):
        """
        NOTE: This is a workaround mechanism, since Cloudera Manager spawns GenerateCredentials commands
        without treating them as child processes or specifying their parent, thus
        this behaviour causes issues when kerberizing via the API.
        :return (True,message) If all unnecessary generate credentials commands have finished appropriately,
        (False, message) otherwise
        """
        all_active_cloudera_manager_resource_commands = self.list_active_cloudera_manager_resource_commands()
        if all_active_cloudera_manager_resource_commands.items:
            all_active_generate_credentials_commands = list(command for command in
                                                            all_active_cloudera_manager_resource_commands.items
                                                            if command.name == "GenerateCredentials")
            for generate_credentials_command in all_active_generate_credentials_commands:
                success, message = self.wait_for_command(generate_credentials_command)
                LOG.info(message)
                if not success:
                    return False, f"An unnecessary GenerateCredentials command did not finish, details: {message} "
            else:
                return True, "All unnecessary GenerateCredentials commands have finished."
        return True, "There were no unnecessary GenerateCredentials commands active"

    def abort_command(self, command: ApiCommand):
        """
        Tries to abort an ApiCommand.

        :param command: The ApiCommand to be aborted, e.g.: GenerateCredentials
        :return: (True,message) If the returned ApiCommand is inactive and unsuccessful, then it's considered aborted
        (False, message) otherwise
        :raises ApiException, If a connection issue occurred
        """
        try:
            api_instance = cm_client.CommandsResourceApi(cm_client.ApiClient())
            api_response = api_instance.abort_command(command.id)
            if not (api_response.active or api_response.success):
                return True, f"Command ID: {command.id}, name: {command.name} successfully aborted."
            else:
                return False, f"Couldn't abort ID: {command.id}, name: {command.name} command"
        except ApiException as e:
            LOG.info(f"Could not read command from server: {e.reason}")
            raise e

    def wait_for_command(self, command: ApiCommand, timeout_sec: int = 360):
        """
        Waits until a command has finished

        :param command An ApiCommand that contains the following useful fields =>
          id: int => The command ID to query
          active: bool => Whether the command is currently active.
          success: bool => If the command is finished, whether it was successful.
          children: ApiCommandList => List of child commands that this command has spawned.
        :param timeout_sec: Maximum timeout in seconds
        :raises ApiException, If a connection issue occurred
        :return: (True , message) if a command and some of spawned processes have succeeded before a timeout
        (False, message) otherwise
        """
        api_instance = cm_client.CommandsResourceApi(cm_client.ApiClient())

        begin_time = time.time()
        timeout = begin_time + timeout_sec
        while time.time() < timeout:
            if command.active:
                time.sleep(10)
                try:  # query command state from API
                    command = api_instance.read_command(command.id)
                except ApiException as e:
                    LOG.info(f"Could not read command from server: {e.reason}")
                    raise e
            else:  # command is inactive
                elapsed_seconds = time.time() - begin_time
                if not command.success:
                    return False, f"Name: {command.name}, ID: {command.id} command has failed." \
                                  f" Elapsed time: {elapsed_seconds:.2f} seconds"
                else:  # command is inactive and has been successful
                    return True, f"Name: {command.name}, ID: {command.id} command has successfully " \
                                 f"finished in {elapsed_seconds:.2f} seconds"
        return False, f"Name: {command.name}, ID: {command.id} command did not finish in {timeout_sec} seconds"

    def update_service_config(self, service_name: str, config_dict: dict):
        LOG.info("Updating configs for '%s' - '%s'" % (service_name, config_dict))
        try:
            api_configs_list = []
            for key, value in config_dict.items():
                api_configs_list.append(cm_client.ApiConfig(name=key, value=value))
            api_service_config = ApiServiceConfig(items=api_configs_list)

            api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
            api_response = api_instance.update_service_config(self.cluster_name, service_name, body=api_service_config)
            LOG.debug("Response %s" % api_response)
            return True, f"Service config for '{service_name}' successfully updated with key-values {config_dict}"
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            return False, str(e)

    def get_current_config_value(self, service_name: str, config_name: str):
        configs = self._read_service_config(service_name)
        for config in configs.items:
            if config.name == config_name:
                return config.value if config.value else config.default
        raise StopIteration

    def _read_service_config(self, service_name: str):
        LOG.info("Getting service configurations for '%s'" % service_name)
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_service_config(self.cluster_name, service_name, view='full')
            LOG.debug("Response %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling ServicesResourceApi->service_command_by_name: %s\n" % e)
            raise e

    def update_cm_config(self, config_dict: dict):
        LOG.info("Attempting to update CM configuration values...")
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        configs = cm_client.ApiConfigList(config_list)

        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.update_config(body=configs)
            LOG.debug("Response %s" % api_response)
            return True, "CM configurations has been updated..."
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->update_config: %s\n" % e)
            return False, str(e)

    def get_all_mgmt_role_config_groups(self) -> ApiRoleConfigGroupList:
        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        return api_instance.read_role_config_groups()

    def update_mgmt_role_config(self, role_config_group_name: str, config_dict: dict):
        LOG.info(f"Attempting to update CM Management '{role_config_group_name}' Role's config...")
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        configs = cm_client.ApiConfigList(config_list)

        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.update_config(role_config_group_name, body=configs)
            LOG.debug("Response %s" % api_response)
            return True, "CM configurations has been updated..."
        except ApiException as e:
            LOG.error("Exception when calling MgmtRoleConfigGroupsResourceApi->update_config: %s\n" % e)
            return False, str(e)

    def check_if_config_is_present_for_mgmt_role_config_group(self, role_config_group_name: str, config_name: str) \
      -> bool:
        api_instance = cm_client.MgmtRoleConfigGroupsResourceApi(cm_client.ApiClient())
        api_response: ApiConfigList = api_instance.read_config(role_config_group_name, view='full')
        return self._check_if_config_is_present(api_response, config_name)

    def _check_if_config_is_present(self, config_list: ApiConfigList, config_name: str):
        config_items: list[ApiConfig] = config_list.items
        try:
            next(conf for conf in config_items if conf.name == config_name)
            return True
        except StopIteration:
            return False

    def read_roles(self, service_name: str):
        LOG.info("Reading roles for service '%s'" % service_name)
        api_instance = cm_client.RolesResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_roles(self.cluster_name, service_name)
            LOG.debug("Response %s" % api_response)
            return api_response
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->read_roles: %s\n" % e)

    def run_role_command(self, command_name: str, service_name: str, role_names: list, timeout_sec: int = DEFAULT_ROLE_COMMAND_TIMEOUT_SECONDS):
        LOG.info("Running '%s' role command for service '%s'" % (command_name, service_name))
        api_instance = cm_client.RoleCommandsResourceApi(cm_client.ApiClient())

        body = cm_client.ApiRoleNameList(role_names)
        try:
            api_response = api_instance.role_command_by_name(self.cluster_name, command_name, service_name, body=body)
            LOG.debug("Response %s" % api_response)
            for role_name in role_names:
                success, message = self.wait_for_role_commands_to_run(role_name, service_name, timeout_sec)
                if not success:
                    return False, message
                else:
                    LOG.info(message)
            else:
                return True, f"Running role command {command_name} on service {service_name} successful"
        except ApiException as e:
            LOG.error("Exception when calling RolesResourceApi->read_roles: %s\n" % e)
            return False, str(e)

    def get_cluster(self, cluster_name: str = None):
        if cluster_name is None:
            cluster_name = self.cluster_name

        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        try:
            # Reads information about a cluster.
            return api_instance.read_cluster(cluster_name)
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->read_cluster: %s\n" % e)
        return None

    def is_cluster_exist(self, cluster_name: str = None):
        return self.get_cluster(cluster_name) is not None

    def is_management_service_exists(self):
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())
        LOG.info(f"Checking whether Cloudera Management Service already exists on cluster {self.cluster_name}, on host"
                 f"{self.hostname}")
        try:
            api_instance.read_service(view='summary')
            return True
        except ApiException:
            return False

    def create_cloudera_management_service(self):
        """
        Attempt to create cloudera management service

        :return: Tuple of ( success , message )
        """
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())
        LOG.info(f"Attempting to create empty management service on cluster {self.cluster_name}, on host"
                 f"{self.hostname}")
        try:
            body = cm_client.ApiService(display_name='Cloudera Management Service')
            api_instance.setup_cms(body=body)
            return True, "Cloudera Management Service successfully created."
        except ApiException as e:
            return False, str(e)

    def update_role_config_group_property(self, service_name: str, role_config_group_display_name: str,
                                          configuration_name: str, configuration_value):
        """
        Attempt to update a property to a given value in a role configuration group of a service

        :return: Tuple of ( success , message )
        """
        try:
            role_config_group_name = self. \
                _get_role_config_group_name_from_all_role_config_groups(service_name, role_config_group_display_name)
            message = f"Updated {service_name} service's {role_config_group_display_name}  role configuration group's" \
                      f" {configuration_name} configuration to the following value: '{configuration_value}'"
            config_to_set = {configuration_name: configuration_value}
            self.update_role_config_group_configuration(service_name, role_config_group_name,
                                                        config_to_set, message)
            return True, message
        except ApiException as e:
            return False, str(e)
        except StopIteration:
            return False, "Couldn't find role config group name among role config groups"

    def get_all_services(self) -> list:
        api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        return api_instance.read_services(self.cluster_name).items

    def get_all_services_silently(self):
        try:
            return True, self.get_all_services()
        except ApiException as e:
            return False, str(e)

    def get_role_config_groups_for_service(self, service_name: str) -> ApiRoleConfigGroupList:
        """
        Get all the role configs groups that are associated to a given service

        :param: service_name
        :return: ApiRoleConfigGroupList
        :except: ApiException
        """
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        return api_instance.read_role_config_groups(self.cluster_name, service_name)

    def check_if_config_is_present_for_role_config_group(
      self, service_name: str, role_config_group_name: str, config_name: str) -> bool:
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        api_response: ApiConfigList = api_instance.read_config(
            self.cluster_name, role_config_group_name, service_name, view='full')
        return self._check_if_config_is_present(api_response, config_name)

    def _get_role_config_group_name_of_role_config_display_name(self, role_config_groups: ApiRoleConfigGroupList,
                                                                display_name: str):
        """
        Find the real role config group name based on the display role config group name

        :param: role_config_groups: The list to search in
        :param: display_name: Filter criteria
        :return: Matching role_config_group.name
        :except: StopException
        """
        return next(role_config_group.name for role_config_group in role_config_groups.items
                    if role_config_group.display_name == display_name)

    def _get_role_config_group_name_from_all_role_config_groups(self, service_name: str, display_name: str):
        role_config_groups = self.get_role_config_groups_for_service(service_name)
        return self._get_role_config_group_name_of_role_config_display_name(role_config_groups,
                                                                            display_name)

    def update_role_config_group_configuration(self, service_name: str, role_config_group_name: str,
                                               config_dict: dict, message: str = ""):
        api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
        config_list = []
        for key, value in config_dict.items():
            config_list.append(cm_client.ApiConfig(name=key, value=value))
        body = cm_client.ApiConfigList(config_list)

        api_instance.update_config(self.cluster_name, role_config_group_name,
                                   service_name, message=message, body=body)

    def read_configuration_from_role_config_group(self, service_name: str, role_config_group_display_name: str,
                                                  configuration_name: str):
        try:
            role_config_group_name = self._get_role_config_group_name_from_all_role_config_groups(service_name,
                                                                                                  role_config_group_display_name)
            api_instance = cm_client.RoleConfigGroupsResourceApi(cm_client.ApiClient())
            api_response = api_instance.read_config(self.cluster_name, role_config_group_name,
                                                    service_name, view='full')
            return True, next(config.value if config.value else config.default
                              for config in api_response.items
                              if config.name == configuration_name)
        except ApiException as e:
            return False, str(e)
        except StopIteration:
            return False, "Couldn't find role config group name among role config groups"

    def is_tls_enabled(self):
        """
        API call to CM server whether TLS is enabled.

        :return:  bool
        :except ApiException
        """
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        return api_instance.is_tls_enabled(self.cluster_name)

    def get_hostnames_belonging_to_role_group(self, role_group_filter: str):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_response = api_instance.list_hosts(self.cluster_name, view='FULL')
            # get hostname from items whose role_ref's 'role_name' contains 'role_group_filter' as a substring
            return True, json.dumps(list(api_host.hostname
                                         for api_host in api_response.items
                                         if any(role_group_filter in role_ref.role_name
                                                for role_ref in api_host.role_refs)))
        except ApiException as e:
            return False, str(e)

    def query_configuration_value(self, service_name: str, configuration_name: str,
                                  configuration_type: str = 'NOT_SPECIFIED',
                                  role_name_filter: str = None):
        class ConfigurationType(Enum):
            SERVICE = partial(self._query_configuration_query_type_is_service, service_name,
                              configuration_name)
            ROLE = partial(self._query_configuration_query_type_is_role, service_name,
                           configuration_name, role_name_filter)
            NOT_SPECIFIED = partial(self._query_configuration_query_type_is_not_specified, service_name,
                                    configuration_name, role_name_filter)

        try:
            query_function = ConfigurationType[configuration_type].value
            return query_function()
        except ApiException as e:
            return False, str(e)

    def _query_configuration_query_type_is_not_specified(self, service_name, configuration_name: str,
                                                         role_name_filter: str):
        LOG.info("Query type not specified, querying from service configurations first")
        success, configuration_value = self. \
            _query_configuration_query_type_is_service(service_name, configuration_name)

        if success:
            return True, configuration_value

        LOG.info("Trying to query role configurations next")
        return self._query_configuration_query_type_is_role(service_name, configuration_name, role_name_filter)

    def _query_configuration_query_type_is_role(self, service_name, configuration_name,
                                                role_name_filter: str = None):
        api_instance = cm_client.RolesResourceApi(cm_client.ApiClient())
        api_response = api_instance.read_roles(self.cluster_name, service_name)

        if role_name_filter:
            all_roles_for_service = list(item for item in api_response.items if role_name_filter in item.name)
        else:
            all_roles_for_service = api_response.items
        for current_role in all_roles_for_service:
            role_config_api_response = api_instance.read_role_config(self.cluster_name, current_role.name, service_name,
                                                                     view='full')
            try:
                matching_configuration = next(configuration for configuration in role_config_api_response.items
                                              if configuration.name == configuration_name)
                LOG.info(f"Found {configuration_name} in role name: {current_role.name}")
                return True, matching_configuration.value if matching_configuration.value is not None \
                    else matching_configuration.default
            except StopIteration:
                pass
        else:
            return False, f"Couldn't find {configuration_name} in any of the roles for {service_name}"

    def _query_configuration_query_type_is_service(self, service_name, configuration_name):
        service_api_instance = cm_client.ServicesResourceApi(cm_client.ApiClient())
        new_api_response = service_api_instance.read_service_config(self.cluster_name, service_name,
                                                                    view='full')
        try:
            service_configuration = next(config for config in new_api_response.items
                                         if config.name == configuration_name)
            return True, service_configuration.value if service_configuration.value is not None \
                else service_configuration.default
        except StopIteration:
            return False, f"Couldn't find '{configuration_name}' in the service configuration of '{service_name}'"

    def import_kerberos_admin_credentials(self, username: str, password: str):
        LOG.info("Attempting to import kerberos admin credentials to CM...")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            success, message = self.wait_for_unnecessary_generate_credentials_commands()
            if success:
                api_response = api_instance.import_admin_credentials(username=username, password=password)
                LOG.debug(f"Response: {api_response}")
                return True, "Imported kerberos admin credentials to CM"
            else:
                return False, message
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->import_admin_credentials: %s\n" % e)
            return False, str(e)

    def configure_cluster_for_kerberos(self):
        LOG.info("Calling configure for kerberos API endpoint on CM...")
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        body = self.create_configure_kerberos_argument_body()

        try:
            success, message = self.wait_for_unnecessary_generate_credentials_commands()
            if success:
                api_response = api_instance.configure_for_kerberos(self.cluster_name, body=body)
                return self.wait_for_command(api_response)
            else:
                return False, message
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->configure_for_kerberos: %s\n" % e)
            return False, str(e)

    def create_configure_kerberos_argument_body(self):
        success_transceiver_port, response_transceiver_port = \
            self.read_configuration_from_role_config_group("hdfs", "hdfs-DATANODE", "dfs_datanode_port");
        success_datanode_web_port, response_datanode_web_port = \
            self.read_configuration_from_role_config_group("hdfs", "hdfs-DATANODE", "dfs_datanode_http_port");
        # since ApiConfigureForKerberosArguments doesn't correctly fill out default values,
        # we'll have to do it manually
        datanode_transceiver_port = int(response_transceiver_port) if success_transceiver_port else 1004
        datanode_web_port = int(response_datanode_web_port) if success_datanode_web_port else 1006
        return cm_client.ApiConfigureForKerberosArguments(datanode_transceiver_port=datanode_transceiver_port,
                                                          datanode_web_port=datanode_web_port)

    def delete_kerberos_credentials(self, delete_credentials_mode='all'):
        LOG.info(f"Calling Kerberos credential deletion with mode: {delete_credentials_mode}")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            success, message = self.wait_for_unnecessary_generate_credentials_commands()
            if success:
                api_response = api_instance.delete_credentials_command(delete_credentials_mode=delete_credentials_mode)
                LOG.debug(f"Response: {api_response}")
                return self.wait_for_command(api_response)
            else:
                return False, message
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->delete_credentials_command: %s\n" % e)
            return False, str(e)

    def generate_missing_kerberos_credentials(self):
        LOG.info("Initiation of Kerberos credential generation...")
        api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())

        try:
            success, message = self.wait_for_unnecessary_generate_credentials_commands()
            if success:
                api_response = api_instance.generate_credentials_command()
                LOG.debug(f"Response: {api_response}")
                return self.wait_for_command(api_response)
            else:
                return False, message
        except ApiException as e:
            LOG.error("Exception when calling ClouderaManagerResourceApi->generate_credentials_command: %s\n" % e)
            return False, str(e)

    def read_cm_management_service(self) -> ApiService:
        LOG.info("Read CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            return api_instance.read_service()
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->read_service: %s\n" % e)
            raise e

    def stop_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating STOP command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.stop_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STOPPED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->stop_command: %s\n" % e)
            return False, str(e)

    def start_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating START command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.start_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->start_command: %s\n" % e)
            return False, str(e)

    def restart_cm_management_service(self, timeout_sec: int = 360):
        LOG.info("Initiating RESTART command for CM Management Service")
        api_instance = cm_client.MgmtServiceResourceApi(cm_client.ApiClient())

        try:
            api_response = api_instance.restart_command()
            LOG.debug(f"Response: {api_response}")
            expected_service_state = {'service_state': 'STARTED',
                                      'health_summary': '!DISABLED'}
            return self.wait_for_service_state('mgmt', expected_service_state, timeout_sec)
        except ApiException as e:
            LOG.error("Exception when calling MgmtServiceResourceApi->restart_command: %s\n" % e)
            return False, str(e)

    def enable_tls_for_services(self):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_instance.configure_auto_tls_services_command(self.cluster_name)
            return True, f"Successfully enabled auto TLS for services in cluster '{self.cluster_name}'!"
        except ApiException as e:
            return False, str(e)

    def enable_tls(self, remote_username: str, **kwargs):
        """
        Enabling TLS in CM. No Root CA or custom cert is given, CM generates everything in this case.

        :param remote_username: The username to pass to CM server so that it can deploy the certificates to agents
        :param kwargs: can be 'private_key_file' or 'password'
        :return: tuple of success, message
        """
        try:
            api_instance = cm_client.ClouderaManagerResourceApi(cm_client.ApiClient())
            if kwargs.get('private_key_file'):
                text_file = open(kwargs.get('private_key_file'))
                private_key = text_file.read()
                text_file.close()
                body = cm_client.ApiGenerateCmcaArguments(user_name=remote_username, private_key=private_key)
            elif 'password' in kwargs:
                body = cm_client.ApiGenerateCmcaArguments(user_name=remote_username, password=kwargs.get('password'))
            else:
                return False, "No private_key_file or password parameter. Please supply one or the other. If both" \
                              "is supplied, private_key_file takes precedence."
            api_instance.generate_cmca(body=body)

            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_instance.configure_auto_tls_services_command(self.cluster_name)
            return True, f"Successfully enabled auto TLS for services in cluster '{self.cluster_name}'!"
        except ApiException as e:
            return False, str(e)

    def start_all_services(self):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_response = api_instance.start_command(self.cluster_name)

            if api_response.success is not None and not api_response.success:
                return api_response.active, api_response.result_message

            return self.wait_until_all_services_are_started()
        except ApiException as e:
            return False, str(e)

    def stop_all_services(self):
        try:
            api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
            api_response = api_instance.stop_command(self.cluster_name)

            if api_response.success is not None and not api_response.success:
                return api_response.active, api_response.result_message

            return self.wait_for_command(api_response)
        except ApiException as e:
            return False, str(e)

    def wait_until_all_services_are_started(self):
        try:
            all_services = self.get_all_services()
        except ApiException as e:
            return False, f"Couldn't read services from API {str(e)}"
        expected_service_state = {'service_state': 'STARTED', 'health_summary': '!DISABLED'}
        all_services = list(service for service in all_services if service.service_state != 'NA')
        for service in all_services:
            is_service_started, message = self.wait_for_service_state(service.name, expected_service_state)
            if not is_service_started:
                return is_service_started, message
        return True, "All services have been successfully started"

    def deploy_cluster_client_config(self):
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        body = cm_client.ApiHostRefList(items=[])

        try:
            success, message = self.wait_for_unnecessary_generate_credentials_commands()
            if success:
                api_response = api_instance.deploy_cluster_client_config(self.cluster_name, body=body)
                return self.wait_for_command(api_response)
            else:
                return False, message
        except ApiException as e:
            return False, str(e)

    def read_configuration_from_host(self, host_id: str, configuration_name: str):
        api_instance = cm_client.HostsResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.read_host_config(host_id, view='full')
            configuration = next(api_config for api_config in api_response.items
                                 if api_config.name == configuration_name)

            if configuration.value is not None:
                return True, configuration.value

            if configuration.default is not None:
                return True, configuration.default

            return False, f"Couldn't get either configuration value or default " \
                          f"for configuration {configuration_name} on host {host_id}"
        except ApiException as e:
            return False, str(e)
        except StopIteration:
            return False, f"{configuration_name} not found for {host_id}. Available configurations: " + \
                   ", ".join(list(api_config.name for api_config in api_response.items))

    def list_active_cluster_commands(self):
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        return api_instance.list_active_commands(self.cluster_name, view='full')

    def wait_for_deploy_client_configs_and_refresh(self):
        return self.wait_for_commands_to_run(self.list_active_cluster_commands, "DeployClientConfigsAndRefresh")

    def deploy_client_configs_and_refresh(self):
        api_instance = cm_client.ClustersResourceApi(cm_client.ApiClient())
        try:
            api_response = api_instance.deploy_client_configs_and_refresh(self.cluster_name)
            LOG.debug("Response: %s" % api_response)
            return self.wait_for_deploy_client_configs_and_refresh()
        except ApiException as e:
            LOG.error("Exception when calling ClustersResourceApi->deploy_client_configs_and_refresh: %s\n" % e)
            return False, str(e)

